import osiotk as os
from .pypipackage import PYPIPackage as _PYPIPackage


def __build_dir(__basedir: str, subdir_structure, overwrite: bool = False):
    def _ensure_file(__name: str, is_abspath: bool = False, overwrite: bool = False):
        path = os.abspath(__name, is_abspath)
        _exists = os.file_exists(path, is_abspath=True)
        if (not _exists) or (_exists and overwrite):
            os.writes(path, is_abspath=True, content="")

    path = os.abspath(__basedir)
    os.mkdir(path)

    if subdir_structure is not None:
        if isinstance(subdir_structure, list) or isinstance(subdir_structure, tuple):
            for subfile in subdir_structure:
                if isinstance(subfile, str):
                    subpath = os.join_paths(path, subfile)
                    _ensure_file(subpath, is_abspath=True, overwrite=overwrite)
                elif isinstance(subfile, dict):
                    for key, value in subfile.items():
                        p = os.join_paths(path, key)
                        __build_dir(p, subdir_structure=value)
        elif isinstance(subdir_structure, str):
            subpath = os.join_paths(path, subdir_structure)
            _ensure_file(subpath, is_abspath=True, overwrite=overwrite)

        elif isinstance(subdir_structure, dict):
            for key, value in subdir_structure.items():
                p = os.join_paths(path, key)
                __build_dir(p, subdir_structure=value)
        else:
            print("unable to make folder structure:", subdir_structure)


def build_package():

    package = _PYPIPackage()
    if package is not None:
        __build_dir(package.package_dir, subdir_structure=package.subdir_structure)

        for key in ("pyproject_toml", "readme", "private"):
            path = getattr(package.paths, key)
            content = getattr(package.content, key)

            os.writes(path, content=content, is_abspath=True)



from .util import get_package_dir as _get_package_dir
from .paths import Paths
from .config import Config, init_config as _init_config
from .content import Content


class PYPIPackage:
    def __init__(self, __config: Config = None):
        self.config = _init_config() if __config is None else __config

    @property
    def package_dir(self):
        return _get_package_dir(self.config)

    @property
    def package_name(self):
        return self.config["package_name"]

    @property
    def subdir_structure(self):
        return [
            "pyproject.toml",
            "MANIFEST.in",
            "README.md",
            "private.txt",
            {"tests": ["__init__.py", f"test_{self.package_name}.py"]},
            {"src": ["__init__.py", {self.package_name: ["__init__.py"]}]},
        ]

    @property
    def paths(self):
        return Paths(self.package_dir)

    @property
    def content(self):
        return Content(self.config)



def get_package_dir(__config: dict):
    package_dir = str(__config["package_dir"])
    package_name = str(__config["package_name"])
    while package_dir.endswith("/") and len(package_dir) > 1:
        package_dir = package_dir[:-1]
    if package_dir.endswith(package_name):
        package_dir += "/"
    else:
        package_dir += f"/{package_name}"

    if package_dir.startswith("~"):
        package_dir = package_dir[1:]
    if not package_dir.startswith("/"):
        package_dir = package_dir + "/"
    return package_dir



from typing import List
from .config import Config


def _form_command(__name: str, command: str):
    return f"# {__name.title()}:\n#    {command}"


def _filter_if_exists(__iter):
    return (value for value in iter(__iter) if value)


class Command:
    def __init__(self, __name: str, public: str, private: str = ""):
        self.name = __name
        self.public = public
        self.private = private if private else public

    @property
    def public_command(self):
        return _form_command(self.name, self.public) if self.public else ""

    @property
    def private_command(self):
        return _form_command(self.name, self.private) if self.private else ""


class Commands(List[Command]):
    @property
    def public_commands(self):
        return _filter_if_exists(command.public_command for command in self)

    @property
    def private_commands(self):
        return _filter_if_exists(command.private_command for command in self)


def commands(__config: Config):

    install_poetry = Command("install poetry", public="python3 -m pip install poetry")

    build = Command("build package", public="python3 -m poetry build")

    pypi_username = __config.pypi_username
    package_name = __config.package_name
    pypi_password = __config.pypi_password
    package_version = __config.package_version

    publish = Command(
        "publish package",
        public=f"python3 -m poetry publish -u <username> -p <password>",
        private=f"python3 -m poetry publish -u {pypi_username} -p {pypi_password}",
    )

    install_locally = Command(
        "install locally",
        public=f"python3 -m pip install {package_name}=={package_version}",
    )

    build_and_publish = Command(
        "build and publish package",
        public=f"{build.public};{publish.public}",
        private=f"{build.public};{publish.private}",
    )

    build_publish_and_install = Command(
        "build package; publish package; install package",
        public=f"{build_and_publish.public};{install_locally.public}",
        private=f"{build_and_publish.private};{install_locally.public}",
    )

    return Commands(
        (
            install_poetry,
            build,
            publish,
            install_locally,
            build_and_publish,
            build_publish_and_install,
        )
    )



from .commands import commands as _init_commands
from .templates import pyproject_toml as _pyproject_toml_template


class Content:
    def __init__(self, __config):
        self.config = __config

    @property
    def _commands(self):
        return _init_commands(self.config)

    @property
    def _private_commands(self):
        return self._commands.private_commands

    @property
    def _public_commands(self):
        return self._commands.public_commands

    @property
    def pyproject_toml(self) -> str:
        result = _pyproject_toml_template
        for key, value in self.config.items():
            result = result.replace(f"*{key}*", value)
        return result

    @property
    def readme(self) -> str:
        name = str(self.config["package_name"]).title()
        description = str(self.config["package_description"])
        commands = "\n\n".join(self._public_commands)
        return f"## {name}\n\n## Description:\n#    {description}\n\n## Commands:\n\n{commands}\n"

    @property
    def private(self):
        commands = "\n\n".join(self._private_commands)
        return f"## Commands:\n\n {commands}"



pyproject_toml = """
[tool.poetry]
name = "*package_name*"
version = "*package_version*"
description = "*package_description*"
authors = ["*pypi_username* <*pypi_email*>"]
license = "MIT"
readme = "README.md"

[tool.poetry.dependencies]
python = "^*python_version*"


[build-system]
requires = ["poetry-core"]
build-backend = "poetry.core.masonry.api"
"""



import osiotk as _os


class Paths:
    def __init__(self, package_dir: str):
        self.package_dir = package_dir

    @property
    def pyproject_toml(self) -> str:
        return _os.join_paths(self.package_dir, "pyproject.toml")

    @property
    def readme(self) -> str:
        return _os.join_paths(self.package_dir, "README.md")

    @property
    def private(self) -> str:
        return _os.join_paths(self.package_dir, "private.txt")



import osiotk as os
import sys as _sys
from dataclasses import dataclass, field, asdict


@dataclass(frozen=True, order=True)
class _SystemArg:

    index: int = field(init=True)
    arg: str = field(init=True)

    @property
    def argname(self):
        return self.arg.split("=", 1)[0] if "=" in self.arg else ""

    @property
    def argvalue(self):
        return self.arg.split("=", 1)[1] if "=" in self.arg else self.arg


class _SystemArgs(list[_SystemArg]):
    def __contains__(self, __key: str):
        for arg in self:
            if arg.argname == __key:
                result = True
                break
        else:
            result = False
        return result

    def __getitem__(self, __key: str | int):
        if isinstance(__key, str):
            for arg in iter(self):
                if arg.argname == __key:
                    result = arg
                    break
            else:
                result = None
        else:
            result = list.__getitem__(self, __key)
        return result

    def __delitem__(self, __key: str | int):
        if isinstance(__key, str):
            other = self.copy()
            for (i, arg) in enumerate(iter(other)):
                if arg.argname == __key:
                    list.__delitem__(self, i)
                    break

        else:
            list.__delitem__(self, __key)

    def getstr(self, __key: str) -> (str | None):
        result = self[__key]
        return None if result is None else str(result.argvalue)


def _system_args():
    return _SystemArgs(
        _SystemArg(index=index, arg=arg) for (index, arg) in enumerate(_sys.argv)
    )


@dataclass(frozen=False, order=True)
class Config:

    package_name: str = field(init=True)
    package_dir: str = field(init=True)
    pypi_username: str = field(init=True)
    pypi_password: str = field(init=True)
    pypi_email: str = field(init=True)
    package_version: str = field(init=True)
    package_description: str = field(init=True)
    python_version: str = field(init=True)
    license: str = field(init=True)

    def __getitem__(self, key):
        return getattr(self, key)

    def __contains__(self, key):
        return True if self[key] else False

    def __setitem__(self, __key, value):
        setattr(self, __key, value)

    def items(self):
        return asdict(self).items()

    @classmethod
    def fields(cls):
        return tuple(cls.__dataclass_fields__.keys())


__config_defaults = {
    "package_name": "pypi_package_defaultname",
    "package_dir": "/documents/pypi_packages/",
    "pypi_username": "pypi_username",
    "pypi_password": "pypi_password",
    "pypi_email": "pypi_email",
    "package_version": "1.0.0",
    "package_description": "package_description",
    "python_version": "3.7",
    "license": "MIT",
}


def __read_config_file(__path: str):
    kwargs = {}
    config_text = os.reads(__path, is_abspath=True)
    for line in (
        line for line in config_text.splitlines(keepends=False) if "=" in line
    ):
        parts = line.split("=", 1)
        k, v = parts[0], parts[1]
        kwargs[k] = v
    return kwargs


def __load_config(system_args, config: dict = {}):

    print("CONFIG:", config)
    for config_key in Config.fields():
        if not config_key in config:
            if system_args[config_key] is None:
                value = __config_defaults[config_key]
                message = f"Enter value for {config_key}:: (default_value = {value})\n:"
                user_value = input(message)
                if user_value:
                    value = user_value
                config[config_key] = value
    return config


def __approve_config(config: dict):
    response_str = "\n".join(f"    {k}:{v}" for k, v in config.items())
    message = f"\nBuilding package with config:\n\n{response_str}\n\ncontinue? (y/n)\n:"
    approved = input(message).lower()
    config = config if approved == "y" else None


def init_config():
    config = {}
    system_args = _system_args()
    config_path = system_args.getstr("config_path")
    if config_path is not None:
        config = __read_config_file(config_path)
        del system_args["config_path"]
    else:
        __load_config(system_args=system_args, config=config)
    __approve_config(config=config)
    if config is not None and config:
        config = Config(**config)

    return config
